//for CSC 471 - texture mapping lab
//ZJ WOOD

#ifdef __APPLE__
#include "GLUT/glut.h"
#include <OPENGL/gl.h>
#endif
#ifdef __unix__
#include <GL/glut.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include "GLSL_helper.h"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp" //perspective, trans etc
#include "glm/gtc/type_ptr.hpp" //value_ptr
#include "mesh.h"
#include "GeometryCreator.h"
#include "MStackHelp.h"

#include <fstream>
#include <iostream>
#include <math.h>
#include <assert.h>
#include <vector>

#define PI 3.14159265
#define SHIP_Z -5.5
#define SPEED 0.02
#define NUM_BUILDINGS 5
#define DRAG_PARTS 7
#define DRAG_AMP 15
#define DRAG_DIFF 45
#define DRAG_COLL 8
#define LASER_SPEED 0.2
#define HEALTH 5

using namespace std;
using namespace glm;

/*data structure for the image used for  texture mapping */
typedef struct Image {
  unsigned long sizeX;
  unsigned long sizeY;
  char *data;
} Image;

Image *TextureImage;

typedef struct Dragon {
   vec3 pos;
   struct BoundingBox bound[DRAG_COLL];
   float angle;
   float alpha;
   bool isHit;
} Dragon;

typedef struct Explosion {
   pMesh *exp;
   vec3 pos;
   vec3 col;
   float alpha;
   float scale;
} Explosion;

//forward declaration of image loading and texture set-up code
int ImageLoad(char *filename, Image *image);
GLvoid LoadTexture(char* image_file, int tex_id);

//flag and ID to toggle on and off the shader
int ShadeProg, Shade2;

//Handles to the shader data
GLint h_uTexUnit;
GLint h_uLightPos;
GLint h_uLightColor;
GLint h_uViewPos;
GLint h_aPosition;
GLint h_aTexCoord;
GLint h_uModelMatrix;
GLint h_uViewMatrix;
GLint h_uProjMatrix;
GLint h_uMatAmb, h_uMatDif, h_uMatSpec, h_uMatShine, h_uMatAlpha;
GLuint GrndBuffObj, GIndxBuffObj, GTexBuffObj, GNormBuffObj;
int g_GiboLen;
static float  g_width, g_height;
float camTransX = 0, camTransY = 0;

//===== Read obj file globals ======
int shipNumVert, bldg1NumVert, dragonVert[DRAG_PARTS];
GLuint shipIBO, shipPos, shipTex, shipNorm;
GLuint bldg1IBO, bldg1Pos, bldg1Tex, bldg1Norm;
GLuint dragonIBOH, dragonPosH, dragonTexH, dragonNormH;
GLuint dragonIBOBE, dragonPosBE, dragonTexBE, dragonNormBE;
GLuint dragonIBOBC, dragonPosBC, dragonTexBC, dragonNormBC;
GLuint dragonIBOMS, dragonPosMS, dragonTexMS, dragonNormMS;
GLuint dragonIBOA, dragonPosA, dragonTexA, dragonNormA;
GLuint dragonIBOTS, dragonPosTS, dragonTexTS, dragonNormTS;
GLuint dragonIBOTE, dragonPosTE, dragonTexTE, dragonNormTE;
GLuint h_aNormal;
//===== Ship movement globals ======
float shipX, shipY, shipRoll, shipPitch, shipYaw;
float shipTargetX, shipTargetY;
//===== Laser and explosion globals =====
vector<pMesh*> lasers;
vector<vec3> laserPos, laserVel;
vector<Explosion> explosions;
//===== World movement globals =====
float envzf, envzb;
float envBuildings[NUM_BUILDINGS][3]; // x, z, rotation
float speed = SPEED;
//===== Collision detection globals =====
struct BoundingBox *shipBoundOrig, *bldgBoundOrig, *dragonBoundOrig[DRAG_PARTS];
struct BoundingBox shipBound, bldgBound[NUM_BUILDINGS];
int hitsTaken = 0, dragonsShot = 0;
float hitColor = 0.1;
//===== Dragon globals =====
vector<Dragon> dragons;

//declare a matrix stack
RenderingHelper ModelTrans;

static const float g_groundY = -4;      // y coordinate of the ground
static const float g_groundSize = 50.0;   // half the ground length

/* Update a bounding box if the object is transformed */
void updateBoundingBox(struct BoundingBox *orig, struct BoundingBox *updated, mat4 trans) {
   float swap;

   updated->max = trans * orig->max;
   updated->min = trans * orig->min;
   
   if (updated->min.x > updated->max.x) {
      swap = updated->min.x;
      updated->min.x = updated->max.x;
      updated->max.x = swap;
   }
   if (updated->min.y > updated->max.y) {
      swap = updated->min.y;
      updated->min.y = updated->max.y;
      updated->max.y = swap;
   }
   if (updated->min.z > updated->max.z) {
      swap = updated->min.z;
      updated->min.z = updated->max.z;
      updated->max.z = swap;
   }
}

/* helper function to set up material for shading */
void SetMaterial(int i, float newAlpha = -1, vec3 newCol = vec3(-1, -1, -1)) {
   glUseProgram(ShadeProg);
   switch (i) {
      case 0: // Ground
         safe_glUniform3f(h_uMatDif, -1.0, -1.0, -1.0);
         safe_glUniform3f(h_uMatSpec, 0.0, 0.0, 0.0);
         safe_glUniform3f(h_uMatAmb, 0.05, 0.05, 0.05);
         safe_glUniform1f(h_uMatShine, 1.0);
         safe_glUniform1f(h_uMatAlpha, 1.0);
         break; 
      case 1: // Spaceship
         safe_glUniform3f(h_uMatDif, -1.0, -1.0, -1.0);
         safe_glUniform3f(h_uMatSpec, 0.5, 0.5, 0.5);
         safe_glUniform3f(h_uMatAmb, 0.1, 0.1, 0.15);
         if (newCol[0] != -1)
            safe_glUniform3f(h_uMatAmb, newCol[0], newCol[1], newCol[2]);
         else
            safe_glUniform3f(h_uMatAmb, 0.1, 0.1, 0.15);
         safe_glUniform1f(h_uMatShine, 10.0);
         safe_glUniform1f(h_uMatAlpha, 1.0);
         break; 
      case 2: // Building
         safe_glUniform3f(h_uMatDif, -2.0, -2.0, -2.0);
         safe_glUniform3f(h_uMatSpec, 0.0, 0.0, 0.0);
         safe_glUniform3f(h_uMatAmb, 0.05, 0.05, 0.05);
         safe_glUniform1f(h_uMatShine, 1.0);
         safe_glUniform1f(h_uMatAlpha, 1.0);
         break; 
      case 3: // Dragon
         safe_glUniform3f(h_uMatDif, 0.8, 0.0, 0.0);
         safe_glUniform3f(h_uMatSpec, 0.9, 0.9, 0.9);
         safe_glUniform3f(h_uMatAmb, 0.2, 0.1, 0.1);
         safe_glUniform1f(h_uMatShine, 100.0);
         if (newAlpha != -1)
            safe_glUniform1f(h_uMatAlpha, newAlpha);
         else
            safe_glUniform1f(h_uMatAlpha, 1.0);
         break; 
      case 4: // Laser
         safe_glUniform3f(h_uMatDif, 0.0, 0.0, 0.0);
         safe_glUniform3f(h_uMatSpec, 0.0, 0.0, 0.0);
         safe_glUniform3f(h_uMatAmb, 0.2, 0.6, 1.0);
         safe_glUniform1f(h_uMatShine, 1.0);
         safe_glUniform1f(h_uMatAlpha, 1.0);
         break; 
      case 5: // Explosion
         safe_glUniform3f(h_uMatDif, 0.0, 0.0, 0.0);
         safe_glUniform3f(h_uMatSpec, 0.0, 0.0, 0.0);
         if (newCol[0] != -1)
            safe_glUniform3f(h_uMatAmb, newCol[0], newCol[1], newCol[2]);
         else
            safe_glUniform3f(h_uMatAmb, 1.0, 0.0, 0.0);
         safe_glUniform1f(h_uMatShine, 1.0);
         if (newAlpha != -1)
            safe_glUniform1f(h_uMatAlpha, newAlpha);
         else
            safe_glUniform1f(h_uMatAlpha, 1.0);
         break; 
   }
}

/* projection matrix */
void SetProjectionMatrix() {
   glm::mat4 Projection = glm::perspective(80.0f, (float)g_width/g_height, 0.1f, 90.f);
   safe_glUniformMatrix4fv(h_uProjMatrix, glm::value_ptr(Projection));
}

/* camera controls */
void SetView() {
   glm::mat4 Trans = glm::translate( glm::mat4(1.0f), glm::vec3(camTransX, camTransY, 0));
   safe_glUniformMatrix4fv(h_uViewMatrix, glm::value_ptr(Trans));
}

/* set the model transform to the identity */
void SetModelI() {
   glm::mat4 tmp = glm::mat4(1.0f);
   safe_glUniformMatrix4fv(h_uModelMatrix, glm::value_ptr(tmp));
}

/* Remove the view transform in order to set the on screen text in a single place */
void UnSetView() {
   safe_glUniformMatrix4fv(h_uViewMatrix, glm::value_ptr(mat4(1.0f)));
}

void SetShipModel(float dx, float dy, float dz, float sca, float roll, float pitch, float yaw) {   
   mat4 Scale = scale(mat4(1.0f), vec3(sca));
   mat4 Rotate = rotate(mat4(1.0f), yaw, vec3(0, 1, 0)) * rotate(mat4(1.0f), pitch, vec3(1, 0, 0)) *
                 rotate(mat4(1.0f), roll, vec3(0, 0, 1)) * rotate(mat4(1.0f), -90.f, vec3(0, 1, 0));
   mat4 Translate = translate(mat4(1.0f), vec3(dx, dy, dz));
   mat4 Trans = Translate * Rotate * Scale;
   safe_glUniformMatrix4fv(h_uModelMatrix, value_ptr(Trans));
   
   updateBoundingBox(shipBoundOrig, &shipBound, Trans);
}

void SetEnvModel(float x, float y, float z, int bldgNum = -1, float sca = 1, float rot = 0) {
   mat4 Rotate = rotate(mat4(1.0f), rot, vec3(0, 1, 0));
   mat4 Scale = scale(mat4(1.0f), vec3(sca));
   mat4 Translate = translate(mat4(1.0f), vec3(x, y, z));
   mat4 Trans = Translate * Rotate * Scale;
   safe_glUniformMatrix4fv(h_uModelMatrix, value_ptr(Trans));
   
   if (bldgNum >= 0 && bldgNum < NUM_BUILDINGS)
      updateBoundingBox(bldgBoundOrig, &bldgBound[bldgNum], Trans);
}

void SetStackModel() {
   safe_glUniformMatrix4fv(h_uModelMatrix, value_ptr(ModelTrans.modelViewMatrix));
}

static void initGround() {

   // A x-z plane at y = g_groundY of dimension [-g_groundSize, g_groundSize]^2
   float GrndPos[] = {
      -g_groundSize, g_groundY, -g_groundSize,
      -g_groundSize, g_groundY,  g_groundSize,
      g_groundSize, g_groundY,  g_groundSize,
      g_groundSize, g_groundY, -g_groundSize
   };

   static GLfloat GrndNorm[] = {
      0, 1, 0,
      0, 1, 0,
      0, 1, 0,
      0, 1, 0
   };

   static GLfloat GrndTex[] = {
      0, 0,
      0, 1,
      1, 1,
      1, 0
   }; 

   unsigned short idx[] = {0, 1, 2, 0, 2, 3};

   g_GiboLen = 6;
   glGenBuffers(1, &GrndBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, GrndBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(GrndPos), GrndPos, GL_STATIC_DRAW);

   glGenBuffers(1, &GIndxBuffObj);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, GIndxBuffObj);
   glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idx), idx, GL_STATIC_DRAW);

   glGenBuffers(1, &GTexBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, GTexBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(GrndTex), GrndTex, GL_STATIC_DRAW);
   
   glGenBuffers(1, &GNormBuffObj);
   glBindBuffer(GL_ARRAY_BUFFER, GNormBuffObj);
   glBufferData(GL_ARRAY_BUFFER, sizeof(GrndNorm), GrndNorm, GL_STATIC_DRAW);
}

void initObj() {
   Model model1, model2;
   Model md0, md1, md2, md3, md4, md5, md6;
   ifstream modelFile;
   
   modelFile.open("Models/space_frigate_6/space_frigate_6.obj");
   model1.load(modelFile);
   shipBoundOrig = (struct BoundingBox *)malloc(sizeof(struct BoundingBox));
   model1.getBoundingBox(shipBoundOrig);
   shipNumVert = model1.meshes()[0].makeVBO(&shipIBO, &shipPos, &shipTex, &shipNorm);
   modelFile.close();
   
   modelFile.open("Models/building1/building1.obj");
   model2.load(modelFile);
   bldgBoundOrig = (struct BoundingBox *)malloc(sizeof(struct BoundingBox));
   model2.getBoundingBox(bldgBoundOrig);
   bldg1NumVert = model2.meshes()[0].makeVBO(&bldg1IBO, &bldg1Pos, &bldg1Tex, &bldg1Norm);
   modelFile.close();
   
   modelFile.open("Models/dragon/head.obj");
   md0.load(modelFile);
   dragonBoundOrig[0] = (struct BoundingBox *)malloc(sizeof(struct BoundingBox));
   md0.getBoundingBox(dragonBoundOrig[0]);
   dragonVert[0] = md0.meshes()[0].makeVBO(&dragonIBOH, &dragonPosH, &dragonTexH, &dragonNormH);
   modelFile.close();
   
   modelFile.open("Models/dragon/bodyEnd.obj");
   md1.load(modelFile);
   dragonBoundOrig[1] = (struct BoundingBox *)malloc(sizeof(struct BoundingBox));
   md1.getBoundingBox(dragonBoundOrig[1]);
   dragonVert[1] = md1.meshes()[0].makeVBO(&dragonIBOBE, &dragonPosBE, &dragonTexBE, &dragonNormBE);
   modelFile.close();
   
   modelFile.open("Models/dragon/bodyConnector.obj");
   md2.load(modelFile);
   dragonBoundOrig[2] = (struct BoundingBox *)malloc(sizeof(struct BoundingBox));
   md2.getBoundingBox(dragonBoundOrig[2]);
   dragonVert[2] = md2.meshes()[0].makeVBO(&dragonIBOBC, &dragonPosBC, &dragonTexBC, &dragonNormBC);
   modelFile.close();
   
   modelFile.open("Models/dragon/midsection.obj");
   md3.load(modelFile);
   dragonBoundOrig[3] = (struct BoundingBox *)malloc(sizeof(struct BoundingBox));
   md3.getBoundingBox(dragonBoundOrig[3]);
   dragonVert[3] = md3.meshes()[0].makeVBO(&dragonIBOMS, &dragonPosMS, &dragonTexMS, &dragonNormMS);
   modelFile.close();
   
   modelFile.open("Models/dragon/arm.obj");
   md4.load(modelFile);
   dragonBoundOrig[4] = (struct BoundingBox *)malloc(sizeof(struct BoundingBox));
   md4.getBoundingBox(dragonBoundOrig[4]);
   dragonVert[4] = md4.meshes()[0].makeVBO(&dragonIBOA, &dragonPosA, &dragonTexA, &dragonNormA);
   modelFile.close();
   
   modelFile.open("Models/dragon/tailStart.obj");
   md5.load(modelFile);
   dragonBoundOrig[5] = (struct BoundingBox *)malloc(sizeof(struct BoundingBox));
   md5.getBoundingBox(dragonBoundOrig[5]);
   dragonVert[5] = md5.meshes()[0].makeVBO(&dragonIBOTS, &dragonPosTS, &dragonTexTS, &dragonNormTS);
   modelFile.close();
   
   modelFile.open("Models/dragon/tailEnd.obj");
   md6.load(modelFile);
   dragonBoundOrig[6] = (struct BoundingBox *)malloc(sizeof(struct BoundingBox));
   md6.getBoundingBox(dragonBoundOrig[6]);
   dragonVert[6] = md6.meshes()[0].makeVBO(&dragonIBOTE, &dragonPosTE, &dragonTexTE, &dragonNormTE);
   modelFile.close();
}


void InitGeom() {
   initGround();
   initObj();
   
   int count;
   for (count = 0; count < NUM_BUILDINGS; count++) {
      envBuildings[count][0] = rand() % 30 - 15;
      envBuildings[count][1] = -(rand() % 80) - 20;
      envBuildings[count][2] = rand() % 4 * 90;
   }
   
   ModelTrans.useModelViewMatrix();
   ModelTrans.loadIdentity();
}

/*function to help load the shaders (both vertex and fragment */
int InstallShader(const GLchar *vShaderName, const GLchar *fShaderName) {
   GLuint VS; //handles to shader object
   GLuint FS; //handles to frag shader object
   GLint vCompiled, fCompiled, linked; //status of shader

   VS = glCreateShader(GL_VERTEX_SHADER);
   FS = glCreateShader(GL_FRAGMENT_SHADER);

   //load the source
   glShaderSource(VS, 1, &vShaderName, NULL);
   glShaderSource(FS, 1, &fShaderName, NULL);

   //compile shader and print log
   glCompileShader(VS);
   /* check shader status requires helper functions */
   printOpenGLError();
   glGetShaderiv(VS, GL_COMPILE_STATUS, &vCompiled);
   printShaderInfoLog(VS);

   //compile shader and print log
   glCompileShader(FS);
   /* check shader status requires helper functions */
   printOpenGLError();
   glGetShaderiv(FS, GL_COMPILE_STATUS, &fCompiled);
   printShaderInfoLog(FS);

   if (!vCompiled || !fCompiled) {
      printf("Error compiling either shader %s or %s", vShaderName, fShaderName);
      return 0;
   }

   //create a program object and attach the compiled shader
   ShadeProg = glCreateProgram();
   glAttachShader(ShadeProg, VS);
   glAttachShader(ShadeProg, FS);

   glLinkProgram(ShadeProg);
   /* check shader status requires helper functions */
   printOpenGLError();
   glGetProgramiv(ShadeProg, GL_LINK_STATUS, &linked);
   printProgramInfoLog(ShadeProg);

   glUseProgram(ShadeProg);

   /* get handles to attribute data */
   h_aPosition = safe_glGetAttribLocation(ShadeProg, "aPosition");
   h_aTexCoord = safe_glGetAttribLocation(ShadeProg,  "aTexCoord");
   h_aNormal = safe_glGetAttribLocation(ShadeProg, "aNormal");
   h_uTexUnit = safe_glGetUniformLocation(ShadeProg, "uTexUnit");
   h_uLightPos = safe_glGetUniformLocation(ShadeProg, "uLightPos");
   h_uLightColor = safe_glGetUniformLocation(ShadeProg, "uLightColor");
   h_uViewPos = safe_glGetUniformLocation(ShadeProg, "uViewPos");
   h_uProjMatrix = safe_glGetUniformLocation(ShadeProg, "uProjMatrix");
   h_uViewMatrix = safe_glGetUniformLocation(ShadeProg, "uViewMatrix");
   h_uModelMatrix = safe_glGetUniformLocation(ShadeProg, "uModelMatrix");
   
   h_uMatAmb = safe_glGetUniformLocation(ShadeProg, "uMat.aColor");
   h_uMatDif = safe_glGetUniformLocation(ShadeProg, "uMat.dColor");
   h_uMatSpec = safe_glGetUniformLocation(ShadeProg, "uMat.sColor");
   h_uMatShine = safe_glGetUniformLocation(ShadeProg, "uMat.shine");
   h_uMatAlpha = safe_glGetUniformLocation(ShadeProg, "uMat.alpha");
   
   printf("sucessfully installed shader %d\n", ShadeProg);
   return 1;
}

/* Some OpenGL initialization */
void Initialize ()                  // Any GL Init Code
{
   // Start Of User Initialization
   glClearColor (1.0f, 0.9f, 0.6f, 1.0f);
   // Black Background
   glClearDepth (1.0f);    // Depth Buffer Setup
   glDepthFunc (GL_LEQUAL);    // The Type Of Depth Testing
   glEnable (GL_DEPTH_TEST);// Enable Depth Testing
   /* texture specific settings */
   glEnable(GL_TEXTURE_2D);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}

void DrawDragons(Dragon *drag) {
   int angle = drag->angle;

   ModelTrans.pushMatrix();
      // Head, entire body transformations
      ModelTrans.translate(vec3(drag->pos.x, drag->pos.y + cos(angle * PI/180) * 3, drag->pos.z));
      ModelTrans.scale(0.15, 0.15, 0.15);
      ModelTrans.rotate(0, vec3(0, 1, 0));
      ModelTrans.rotate(sin(angle * PI/180) * DRAG_AMP, vec3(1, 0, 0));
      angle += DRAG_DIFF;
      
      ModelTrans.pushMatrix();
         // Body end (head end)
         ModelTrans.rotate(sin(angle * PI/180) * DRAG_AMP, vec3(1, 0, 0));
         ModelTrans.translate(vec3(0, 0, -8));
         angle += DRAG_DIFF;
         
         ModelTrans.pushMatrix();
            // Body connector (head end)
            ModelTrans.rotate(-sin(angle * PI/180) * DRAG_AMP, vec3(1, 0, 0));
            ModelTrans.translate(vec3(0, 0.75, -12.5));
            angle += DRAG_DIFF;
      
            ModelTrans.pushMatrix();
               // Midsection
               ModelTrans.rotate(-sin(angle * PI/180) * DRAG_AMP, vec3(1, 0, 0));
               ModelTrans.translate(vec3(0, -1.6, -4));
               angle += DRAG_DIFF;
   
               glBindBuffer(GL_ARRAY_BUFFER, dragonPosA);
               safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
               glBindBuffer(GL_ARRAY_BUFFER, dragonNormA);
               safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
               glBindBuffer(GL_ARRAY_BUFFER, dragonTexA);
               safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
               glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, dragonIBOA);
               
               ModelTrans.pushMatrix();
                  // Back left arm
                  ModelTrans.translate(vec3(3, -0.5, -7.5));
                  ModelTrans.rotate(sin(angle * PI/180) * DRAG_AMP, vec3(1, 0, 0));
                  SetStackModel();
                  glDrawElements(GL_TRIANGLES, dragonVert[4], GL_UNSIGNED_INT, 0);
               ModelTrans.popMatrix();
      
               ModelTrans.pushMatrix();
                  // Back right arm
                  ModelTrans.translate(vec3(-3, -0.5, -7.5));
                  ModelTrans.rotate(sin(angle * PI/180) * DRAG_AMP, vec3(1, 0, 0));
                  SetStackModel();
                  glDrawElements(GL_TRIANGLES, dragonVert[4], GL_UNSIGNED_INT, 0);
               ModelTrans.popMatrix();
      
               ModelTrans.pushMatrix();
                  // Front left arm
                  ModelTrans.translate(vec3(3, -0.5, 7.5));
                  ModelTrans.rotate((sin(angle * PI/180) + 1) * DRAG_AMP, vec3(1, 0, 0));
                  ModelTrans.rotate(180, vec3(1, 0, 0));
                  SetStackModel();
                  glDrawElements(GL_TRIANGLES, dragonVert[4], GL_UNSIGNED_INT, 0);
               ModelTrans.popMatrix();
      
               ModelTrans.pushMatrix();
                  // Front right arm
                  ModelTrans.translate(vec3(-3, -0.5, 7.5));
                  ModelTrans.rotate((sin(angle * PI/180) + 1) * DRAG_AMP, vec3(1, 0, 0));
                  ModelTrans.rotate(180, vec3(1, 0, 0));
                  SetStackModel();
                  glDrawElements(GL_TRIANGLES, dragonVert[4], GL_UNSIGNED_INT, 0);
               ModelTrans.popMatrix();
   
               ModelTrans.pushMatrix();
                  // Body connector (tail end)
                  ModelTrans.translate(vec3(0, 1.6, -4));
                  ModelTrans.rotate(sin(angle * PI/180) * DRAG_AMP*1.5, vec3(1, 0, 0));
                  angle += DRAG_DIFF;
      
                  ModelTrans.pushMatrix();
                     // Body end (tail end)
                     ModelTrans.translate(vec3(0, -0.75, -12.5));
                     ModelTrans.rotate(sin(angle * PI/180) * DRAG_AMP*1.5, vec3(1, 0, 0));
                     angle += DRAG_DIFF;
      
                     ModelTrans.pushMatrix();
                        // Tail start
                        ModelTrans.translate(vec3(0, 0, -8));
                        ModelTrans.rotate(sin(angle * PI/180) * DRAG_AMP, vec3(1, 0, 0));
                        angle += DRAG_DIFF;
      
                        ModelTrans.pushMatrix();
                           // Tail end
                           glBindBuffer(GL_ARRAY_BUFFER, dragonPosTE);
                           safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
                           glBindBuffer(GL_ARRAY_BUFFER, dragonNormTE);
                           safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
                           glBindBuffer(GL_ARRAY_BUFFER, dragonTexTE);
                           safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
                           glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, dragonIBOTE);
                           
                           ModelTrans.translate(vec3(0, -2.5, -7));
                           ModelTrans.rotate(-sin(angle * PI/180) * DRAG_AMP*2, vec3(1, 0, 0));
   
                           SetStackModel();
                           glDrawElements(GL_TRIANGLES, dragonVert[6], GL_UNSIGNED_INT, 0);
                           updateBoundingBox(dragonBoundOrig[6], &drag->bound[7], ModelTrans.modelViewMatrix);
                        ModelTrans.popMatrix();
      
                        glBindBuffer(GL_ARRAY_BUFFER, dragonPosTS);
                        safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
                        glBindBuffer(GL_ARRAY_BUFFER, dragonNormTS);
                        safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
                        glBindBuffer(GL_ARRAY_BUFFER, dragonTexTS);
                        safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
                        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, dragonIBOTS);
                        
                        SetStackModel();
                        glDrawElements(GL_TRIANGLES, dragonVert[5], GL_UNSIGNED_INT, 0);
                        updateBoundingBox(dragonBoundOrig[5], &drag->bound[6], ModelTrans.modelViewMatrix);
                     ModelTrans.popMatrix();
      
                     glBindBuffer(GL_ARRAY_BUFFER, dragonPosBE);
                     safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
                     glBindBuffer(GL_ARRAY_BUFFER, dragonNormBE);
                     safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
                     glBindBuffer(GL_ARRAY_BUFFER, dragonTexBE);
                     safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
                     glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, dragonIBOBE);

                     ModelTrans.rotate(180, vec3(0, 1, 0));

                     SetStackModel();
                     glDrawElements(GL_TRIANGLES, dragonVert[1], GL_UNSIGNED_INT, 0);
                     updateBoundingBox(dragonBoundOrig[1], &drag->bound[5], ModelTrans.modelViewMatrix);
                  ModelTrans.popMatrix();
      
                  glBindBuffer(GL_ARRAY_BUFFER, dragonPosBC);
                  safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
                  glBindBuffer(GL_ARRAY_BUFFER, dragonNormBC);
                  safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
                  glBindBuffer(GL_ARRAY_BUFFER, dragonTexBC);
                  safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
                  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, dragonIBOBC);

                  ModelTrans.rotate(180, vec3(0, 1, 0));

                  SetStackModel();
                  glDrawElements(GL_TRIANGLES, dragonVert[2], GL_UNSIGNED_INT, 0);
                  updateBoundingBox(dragonBoundOrig[2], &drag->bound[4], ModelTrans.modelViewMatrix);
               ModelTrans.popMatrix();
      
               glBindBuffer(GL_ARRAY_BUFFER, dragonPosMS);
               safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
               glBindBuffer(GL_ARRAY_BUFFER, dragonNormMS);
               safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
               glBindBuffer(GL_ARRAY_BUFFER, dragonTexMS);
               safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
               glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, dragonIBOMS);
               
               SetStackModel();
               glDrawElements(GL_TRIANGLES, dragonVert[3], GL_UNSIGNED_INT, 0);
               updateBoundingBox(dragonBoundOrig[3], &drag->bound[3], ModelTrans.modelViewMatrix);
            ModelTrans.popMatrix();
      
      
            glBindBuffer(GL_ARRAY_BUFFER, dragonPosBC);
            safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
            glBindBuffer(GL_ARRAY_BUFFER, dragonNormBC);
            safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
            glBindBuffer(GL_ARRAY_BUFFER, dragonTexBC);
            safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, dragonIBOBC);

            SetStackModel();
            glDrawElements(GL_TRIANGLES, dragonVert[2], GL_UNSIGNED_INT, 0);
            updateBoundingBox(dragonBoundOrig[2], &drag->bound[2], ModelTrans.modelViewMatrix);
         ModelTrans.popMatrix();
      
         glBindBuffer(GL_ARRAY_BUFFER, dragonPosBE);
         safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
         glBindBuffer(GL_ARRAY_BUFFER, dragonNormBE);
         safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
         glBindBuffer(GL_ARRAY_BUFFER, dragonTexBE);
         safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
         glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, dragonIBOBE);

         SetStackModel();
         glDrawElements(GL_TRIANGLES, dragonVert[1], GL_UNSIGNED_INT, 0);
         updateBoundingBox(dragonBoundOrig[1], &drag->bound[1], ModelTrans.modelViewMatrix);
      ModelTrans.popMatrix();
      
      glBindBuffer(GL_ARRAY_BUFFER, dragonPosH);
      safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
      glBindBuffer(GL_ARRAY_BUFFER, dragonNormH);
      safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
      glBindBuffer(GL_ARRAY_BUFFER, dragonTexH);
      safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, dragonIBOH);

      SetStackModel();
      glDrawElements(GL_TRIANGLES, dragonVert[0], GL_UNSIGNED_INT, 0);
      updateBoundingBox(dragonBoundOrig[0], &drag->bound[0], ModelTrans.modelViewMatrix);
   ModelTrans.popMatrix();      
}

/* Main display function */
void Draw (void)
{
   int count;
   
   glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   ModelTrans.loadIdentity();
   
   //Start our shader
   glUseProgram(ShadeProg);

   //Set up matrix transforms
   SetProjectionMatrix();
   SetView();
   SetModelI();

   //set up the texture unit
   safe_glUniform1i(h_uTexUnit, 0);
   glEnable(GL_TEXTURE_2D);
   glActiveTexture(GL_TEXTURE0);

   // Set the light position
   safe_glUniform3f(h_uLightPos, 0, 3, 3);
   safe_glUniform3f(h_uLightColor, 1, 1, 0.9);
   safe_glUniform3f(h_uViewPos, -camTransX, -camTransY, 0);

   // Enable back face culling
   glCullFace(GL_BACK);
   glEnable(GL_CULL_FACE);
   
   // Enable transparency
   glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

   // Enable the attributes
   safe_glEnableVertexAttribArray(h_aPosition);
   safe_glEnableVertexAttribArray(h_aTexCoord);
   safe_glEnableVertexAttribArray(h_aNormal);
   
   // Draw the ground planes and loop them
   SetMaterial(0);
   glBindTexture(GL_TEXTURE_2D, 0);

   glBindBuffer(GL_ARRAY_BUFFER, GrndBuffObj);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ARRAY_BUFFER, GTexBuffObj);
   safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0); 
   glBindBuffer(GL_ARRAY_BUFFER, GNormBuffObj);
   safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, GIndxBuffObj);
   
   SetEnvModel(0, 0, envzf);
   glDrawElements(GL_TRIANGLES, g_GiboLen, GL_UNSIGNED_SHORT, 0);
   SetEnvModel(-g_groundSize*2, 0, envzf);
   glDrawElements(GL_TRIANGLES, g_GiboLen, GL_UNSIGNED_SHORT, 0);
   SetEnvModel(g_groundSize*2, 0, envzf);
   glDrawElements(GL_TRIANGLES, g_GiboLen, GL_UNSIGNED_SHORT, 0);
   
   SetEnvModel(0, 0, envzb);
   glDrawElements(GL_TRIANGLES, g_GiboLen, GL_UNSIGNED_SHORT, 0);
   SetEnvModel(-g_groundSize*2, 0, envzb);
   glDrawElements(GL_TRIANGLES, g_GiboLen, GL_UNSIGNED_SHORT, 0);
   SetEnvModel(g_groundSize*2, 0, envzb);
   glDrawElements(GL_TRIANGLES, g_GiboLen, GL_UNSIGNED_SHORT, 0);
   
   if (envzf >= g_groundSize) {
      envzf = envzb;
      envzb = envzb - 2*g_groundSize + 1;
   }

   // Draw the buildings
   SetMaterial(2);
   glBindTexture(GL_TEXTURE_2D, 2);

   glBindBuffer(GL_ARRAY_BUFFER, bldg1Pos);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ARRAY_BUFFER, bldg1Tex);
   safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ARRAY_BUFFER, bldg1Norm);
   safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
   
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bldg1IBO);
   for (count = 0; count < NUM_BUILDINGS; count++) {
      SetEnvModel(envBuildings[count][0], -1, envBuildings[count][1], count, 0.005, envBuildings[count][2]);
      glDrawElements(GL_TRIANGLES, bldg1NumVert, GL_UNSIGNED_INT, 0);
   }

   // Draw the ship
   SetMaterial(1, -1, vec3(hitColor, 0.1, 0.15));
   if (hitColor >= 0.1)
      hitColor -= 0.01;
   glBindTexture(GL_TEXTURE_2D, 1);
   
   glBindBuffer(GL_ARRAY_BUFFER, shipPos);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ARRAY_BUFFER, shipTex);
   safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ARRAY_BUFFER, shipNorm);
   safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, shipIBO);
   
   SetShipModel(shipX, shipY, SHIP_Z, 0.1, shipRoll, shipPitch, shipYaw);
   glDrawElements(GL_TRIANGLES, shipNumVert, GL_UNSIGNED_INT, 0);

   // Draw the dragons
   for (count = 0; count < dragons.size(); count++) {
      dragons[count].pos.z += speed+0.02;
      if (dragons[count].pos.z > 10 || dragons[count].alpha <= 0) {
         dragons.erase(dragons.begin() + count);
         count--;
         continue;
      }
      else if (dragons[count].isHit == true) {
         dragons[count].pos.y -= 0.02;
         dragons[count].alpha -= 0.005;
         SetMaterial(3, dragons[count].alpha);
      }
      else {
         dragons[count].angle += 0.4;
         SetMaterial(3);
      }
      DrawDragons(&dragons[count]);
   }

   // Draw the lasers
   SetMaterial(4);
   for (count = 0; count < lasers.size(); count++) {
      ModelTrans.pushMatrix();
         glBindBuffer(GL_ARRAY_BUFFER, lasers[count]->PositionHandle);
         safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
         glBindBuffer(GL_ARRAY_BUFFER, lasers[count]->NormalHandle);
         safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
         glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, lasers[count]->IndexHandle);
         
         laserPos[count] += laserVel[count];
         if (laserPos[count].z < -90 || laserPos[count].y < g_groundY) {
            lasers.erase(lasers.begin() + count);
            laserPos.erase(laserPos.begin() + count);
            laserVel.erase(laserVel.begin() + count);
            count--;
            continue;
         }
         
         ModelTrans.translate(laserPos[count]);
         ModelTrans.rotate(-atan(laserVel[count].x/-laserVel[count].z) * 180/PI, vec3(0, 1, 0));
         ModelTrans.rotate(atan(laserVel[count].y/-laserVel[count].z) * 180/PI, vec3(1, 0, 0));
         
         SetStackModel();
         glDrawElements(GL_TRIANGLES, lasers[count]->IndexBufferLength, GL_UNSIGNED_SHORT, 0);
      ModelTrans.popMatrix();
   }

   // Draw the explosions
   for (count = 0; count < explosions.size(); count++) {
      ModelTrans.pushMatrix();
         glBindBuffer(GL_ARRAY_BUFFER, explosions[count].exp->PositionHandle);
         safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
         glBindBuffer(GL_ARRAY_BUFFER, explosions[count].exp->NormalHandle);
         safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
         glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, explosions[count].exp->IndexHandle);
         
         explosions[count].pos[2] += speed;
         explosions[count].col[1] += 0.01;
         explosions[count].alpha -= 0.01;
         explosions[count].scale += 0.05;
         if (hitsTaken >= HEALTH)
            explosions[count].alpha += 0.005;
         
         if (explosions[count].alpha <= 0) {
            explosions.erase(explosions.begin() + count);
            count--;
            continue;
         }
         SetMaterial(5, explosions[count].alpha, explosions[count].col);
         ModelTrans.translate(explosions[count].pos);
         ModelTrans.scale(explosions[count].scale, explosions[count].scale, explosions[count].scale);
      
         SetStackModel();
         glDrawElements(GL_TRIANGLES, explosions[count].exp->IndexBufferLength, GL_UNSIGNED_SHORT, 0);
      ModelTrans.popMatrix();
   }

   // Disable the attributes
   safe_glDisableVertexAttribArray(h_aPosition);
   safe_glDisableVertexAttribArray(h_aTexCoord);
   safe_glDisableVertexAttribArray(h_aNormal);

   // Set the on-screen text
   char health[20], kills[20];
   SetMaterial(4);
   sprintf(health, "Health: %d/%d", HEALTH-hitsTaken, HEALTH);
   sprintf(kills, "Kills: %d", dragonsShot);
   UnSetView();
   ModelTrans.pushMatrix();
      ModelTrans.translate(vec3(-4.5, 2.6, -3.5));
      SetStackModel();
   ModelTrans.popMatrix();
   glRasterPos3f(0.0f, 0.0f, 0.0f);
   
   for (count = 0; health[count] != 0; count++)
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, health[count]);

   glRasterPos3f(0.0f, -0.2f, 0.0f);
   
   for (count = 0; kills[count] != 0; count++)
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, kills[count]);

   //Disable the shader
   glUseProgram(0);
   glDisable(GL_TEXTURE_2D);

   glutSwapBuffers();

}

void keyboard(unsigned char key, int x, int y ){
   switch( key ) {
      case 'q': case 'Q' :
         free(shipBoundOrig);
         free(bldgBoundOrig);
         exit( EXIT_SUCCESS );
         break;
   }
   glutPostRedisplay();
}

void passiveMouse(int x, int y) {
   if (hitsTaken < HEALTH) {
      if (x < g_width && x > 0) {
         shipTargetX = (x - 0.5*g_width)/g_height * 14.0;
      }
      if (y < g_height && y > 0) {
         shipTargetY = g_height - y - 1;
         shipTargetY = (shipTargetY / g_height - 0.5) * 18.0;
         if (shipTargetY <= g_groundY + 2)
            shipTargetY = g_groundY + 2;
      }
   }
}

void mouse(int button, int state, int x, int y) {
   if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN && hitsTaken < HEALTH) {
      lasers.push_back(GeometryCreator::CreateCylinder(0.1f, 0.1f, 1.0f, 8, 8));
      laserPos.push_back(vec3(shipX, shipY, SHIP_Z-3));
      laserVel.push_back(vec3(LASER_SPEED * sin(-shipYaw * PI/180), LASER_SPEED * sin(shipPitch * 0.5 * PI/180),
                         -LASER_SPEED * (cos(shipYaw * PI/180) + sin(shipPitch * 0.5 * PI/180)) / 2));
   }
}

void update(int value) {
   static bool isColliding = false;
   static int dragonTimeout = 2000;
   int count, count2, count3;
   bool noCollisions = true;
   Explosion newExp;

   camTransX += -(shipTargetX - shipX) / 180.0;
   camTransY += -(shipTargetY - shipY) / 180.0;
   shipX += (shipTargetX - shipX) / 90.0;
   shipY += (shipTargetY - shipY) / 90.0;
   
   if (hitsTaken < HEALTH) {
      shipRoll = -(shipTargetX - shipX) * 8;
      shipPitch = (shipTargetY - shipY) * 10;
      shipYaw = -(shipTargetX - shipX) * 5;
   }
   else if (shipY > g_groundY+2) {  // Plummeting
      shipRoll += 1;
      shipPitch -= 0.1;
   }
   else if (value > 0) {   // Hit the ground
      newExp.exp = GeometryCreator::CreateSphere(vec3(0.5));
      newExp.pos = vec3(shipX, shipY, SHIP_Z);
      newExp.col = vec3(1, 0, 0);
      newExp.alpha = 1;
      newExp.scale = 1;
      explosions.push_back(newExp);
      value = -2;
   }
   
   if (shipY <= g_groundY+2)
      speed = 0;
   else if (hitsTaken >= HEALTH)
      shipTargetY -= 0.02;
   
   /* Move the objects in the environment */
   envzf += speed;
   envzb += speed;
   
   /* Check for collisions with dragons */
   for (count = 0; count < dragons.size(); count++) {
      for (count2 = 0; count2 < DRAG_COLL; count2++) {      
         for (count3 = 0; count3 < lasers.size(); count3++) {
            if (laserPos[count3].x < dragons[count].bound[count2].max.x && laserPos[count3].x > dragons[count].bound[count2].min.x &&
             laserPos[count3].y < dragons[count].bound[count2].max.y && laserPos[count3].y > dragons[count].bound[count2].min.y &&
             laserPos[count3].z < dragons[count].bound[count2].max.z && laserPos[count3].z > dragons[count].bound[count2].min.z) {               
               newExp.exp = GeometryCreator::CreateSphere(vec3(0.5));
               newExp.pos = laserPos[count3];
               newExp.col = vec3(1, 0, 0);
               newExp.alpha = 1;
               newExp.scale = 1;
               explosions.push_back(newExp);
               
               lasers.erase(lasers.begin() + count3);
               laserPos.erase(laserPos.begin() + count3);
               laserVel.erase(laserVel.begin() + count3);
               count3--;
               
               if (!dragons[count].isHit) {
                  dragons[count].isHit = true;
                  dragonsShot++;
                  if (dragonTimeout - 100 >= 300)
                     dragonTimeout -= 100;
               }
            }
         }
         
         if (noCollisions && dragons[count].bound[count2].max.z > shipBound.min.z - 2 && !dragons[count].isHit) {
            if (!(shipBound.min.x > dragons[count].bound[count2].max.x || dragons[count].bound[count2].min.x > shipBound.max.x ||
                shipBound.min.y > dragons[count].bound[count2].max.y || dragons[count].bound[count2].min.y > shipBound.max.y ||
                shipBound.min.z > dragons[count].bound[count2].max.z || dragons[count].bound[count2].min.z > shipBound.max.z)) {
               noCollisions = false;
               if (!isColliding) {
                  isColliding = true;
                  hitsTaken = hitsTaken + 1 > HEALTH ? hitsTaken : hitsTaken+1;
                  hitColor = 1;
               }
            }
         }
      }
   }
   
   /* Create new dragons */
   if (value > dragonTimeout) {
      Dragon newDragon;
      newDragon.pos = vec3(rand() % 30 - 15, 2, -90);
      for (count = 0; count < DRAG_PARTS; count++)
         newDragon.bound[count] = *dragonBoundOrig[count];
      newDragon.angle = rand() % 360;
      newDragon.alpha = 1.0;
      newDragon.isHit = false;
      dragons.push_back(newDragon);
      value = 0;
   }
   
   /* Update buildings and check for collisions with buildings */
   for (count = 0; count < NUM_BUILDINGS; count++) {
      if (envBuildings[count][1] < 10)
         envBuildings[count][1] += speed;
      else {
         envBuildings[count][0] = rand() % 60 - 30;
         envBuildings[count][1] = -(rand() % 10) - 100;
         envBuildings[count][2] = rand() % 4 * 90;
      }
      
      /* Checking for collisions */
      if (noCollisions && bldgBound[count].max.z > shipBound.min.z - 2) {
         if (!(shipBound.min.x > bldgBound[count].max.x || bldgBound[count].min.x > shipBound.max.x ||
             shipBound.min.y > bldgBound[count].max.y || bldgBound[count].min.y > shipBound.max.y ||
             shipBound.min.z > bldgBound[count].max.z || bldgBound[count].min.z > shipBound.max.z)) {
            noCollisions = false;
            if (!isColliding) {
               isColliding = true;
               hitsTaken = hitsTaken + 1 > HEALTH ? hitsTaken : hitsTaken+1;
               hitColor = 1;
            }
         }
      }
      
      for (count2 = 0; count2 < lasers.size(); count2++) {
         if (laserPos[count2].x < bldgBound[count].max.x && laserPos[count2].x > bldgBound[count].min.x &&
             laserPos[count2].y < bldgBound[count].max.y && laserPos[count2].y > bldgBound[count].min.y &&
             laserPos[count2].z < bldgBound[count].max.z && laserPos[count2].z > bldgBound[count].min.z) {
            newExp.exp = GeometryCreator::CreateSphere(vec3(0.5));
            newExp.pos = laserPos[count2];
            newExp.col = vec3(1, 0, 0);
            newExp.alpha = 1;
            newExp.scale = 1;
            explosions.push_back(newExp);
             
            lasers.erase(lasers.begin() + count2);
            laserPos.erase(laserPos.begin() + count2);
            laserVel.erase(laserVel.begin() + count2);
            count2--;
         }
      }
   }
   
   if (noCollisions)
      isColliding = false;  
   
   glutPostRedisplay();
   glutTimerFunc(1, update, value + 1);
}

/* Reshape */
void ReshapeGL (int width, int height)
{
    g_width = (float)width;
    g_height = (float)height;
    glViewport (0, 0, (GLsizei)(width), (GLsizei)(height));

}

int main(int argc, char** argv) {
   // Set the seed for random numbers
   srand(time(NULL));
  
   //initialize the window
   glutInit(&argc, argv);
   glutInitWindowPosition( 20, 20 );
   glutInitWindowSize(1152, 720);
   glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
   glutCreateWindow("Shoot the dragons!");

   //set up the opengl call backs
   glutDisplayFunc(Draw);
   glutReshapeFunc(ReshapeGL);
   glutKeyboardFunc(keyboard);
   glutMouseFunc(mouse);
   glutMotionFunc(passiveMouse);
   glutPassiveMotionFunc(passiveMouse);
   glutTimerFunc(1, update, 0);

   Initialize();
   //load in the other image textures
   LoadTexture((char *)"Textures/Ground.bmp", 0);
   LoadTexture((char *)"Models/space_frigate_6/space_frigate_6_color.bmp", 1);
   LoadTexture((char *)"Models/building1/building1.bmp", 2);
   

   //test the openGL version
   getGLversion();
   //install the shader
   if (!InstallShader(textFileRead((char *)"spaceship_vert.glsl"), textFileRead((char *)"spaceship_frag.glsl"))) {
      printf("Error installing shader!\n");
      return 0;
   }
   InitGeom();

   shipX = shipY = shipRoll = shipPitch = shipYaw = 0;
   shipTargetX = shipTargetY = 0;
   envzf = -g_groundSize;
   envzb = -3*g_groundSize + 1;
   
   glutMainLoop();
  
}

//routines to load in a bmp files - must be 2^nx2^m and a 24bit bmp
GLvoid LoadTexture(char* image_file, int texID) { 
  
   TextureImage = (Image *) malloc(sizeof(Image));
   if (TextureImage == NULL) {
      printf("Error allocating space for image");
      exit(1);
   }
   cout << "trying to load " << image_file << endl;
   glBindTexture(GL_TEXTURE_2D, texID);
   if (!ImageLoad(image_file, TextureImage)) { /* Load image and create mipmaps */
      exit(1);
   }  
   /*  2d texture, level of detail 0 (normal), 3 components (red, green, blue),            */
   /*  x size from image, y size from image,                                              */    
   /*  border 0 (normal), rgb color data, unsigned byte data, data  */  
   glBindTexture(GL_TEXTURE_2D, texID);
   glTexImage2D(GL_TEXTURE_2D, 0, 3,
                TextureImage->sizeX, TextureImage->sizeY,
                0, GL_RGB, GL_UNSIGNED_BYTE, TextureImage->data);
   glGenerateMipmap(GL_TEXTURE_2D);
}


/* BMP file loader loads a 24-bit bmp file only */

/*
* getint and getshort are help functions to load the bitmap byte by byte
*/
static unsigned int getint(FILE *fp) {
   int c, c1, c2, c3;
  
   /*  get 4 bytes */ 
   c = getc(fp);  
   c1 = getc(fp);  
   c2 = getc(fp);  
   c3 = getc(fp);
  
   return ((unsigned int) c) +   
     (((unsigned int) c1) << 8) + 
     (((unsigned int) c2) << 16) +
     (((unsigned int) c3) << 24);
}

static unsigned int getshort(FILE *fp){
   int c, c1;
  
   /* get 2 bytes*/
   c = getc(fp);  
   c1 = getc(fp);
  
   return ((unsigned int) c) + (((unsigned int) c1) << 8);
}

/*  quick and dirty bitmap loader...for 24 bit bitmaps with 1 plane only.  */

int ImageLoad(char *filename, Image *image) {
   FILE *file;
   unsigned long size;                 /*  size of the image in bytes. */
   unsigned long i;                    /*  standard counter. */
   unsigned short int planes;          /*  number of planes in image (must be 1)  */
   unsigned short int bpp;             /*  number of bits per pixel (must be 24) */
   char temp;                          /*  used to convert bgr to rgb color. */

   /*  make sure the file is there. */
   if ((file = fopen(filename, "rb"))==NULL) {
      printf("File Not Found : %s\n",filename);
      return 0;
   }

   /*  seek through the bmp header, up to the width height: */
   fseek(file, 18, SEEK_CUR);

   /*  No 100% errorchecking anymore!!! */

   /*  read the width */    
   image->sizeX = getint (file);

   /*  read the height */ 
   image->sizeY = getint (file);
   /* Changing sizeY to positive if a negative value was used */
   if ((image->sizeY & 0x80000000) != 0) {
      image->sizeY = (~image->sizeY & 0xFFFFFFFF) + 1;
   }


   /*  calculate the size (assuming 24 bits or 3 bytes per pixel). */
   size = image->sizeX * image->sizeY * 3;

   /*  read the planes */    
   planes = getshort(file);
   if (planes != 1) {
      printf("Planes from %s is not 1: %u\n", filename, planes);
      return 0;
   }

   /*  read the bpp */    
   bpp = getshort(file);
   if (bpp != 24) {
      printf("Bpp from %s is not 24: %u\n", filename, bpp);
      return 0;
   }

   /*  seek past the rest of the bitmap header. */
   fseek(file, 24, SEEK_CUR);

   /*  read the data.  */
   image->data = (char *) malloc(size);
   if (image->data == NULL) {
      printf("Error allocating memory for color-corrected image data");
      return 0; 
   }

   if ((i = fread(image->data, size, 1, file)) != 1) {
      printf("Error reading image data from %s.\n", filename);
      return 0;
   }

   for (i=0;i<size;i+=3) { /*  reverse all of the colors. (bgr -> rgb) */
      temp = image->data[i];
      image->data[i] = image->data[i+2];
      image->data[i+2] = temp;
   }

   fclose(file); /* Close the file and release the filedes */

   /*  we're done. */
   return 1;
}




