struct Material {
  vec3 aColor;
  vec3 dColor;
  vec3 sColor;
  float shine;
  float alpha;
};

uniform sampler2D uTexUnit;
uniform vec3 uLightPos;
uniform vec3 uLightColor;
uniform vec3 uViewPos;
uniform Material uMat;

varying vec2 vTexCoord;
varying vec3 vNormal;
varying vec3 vPos;

void main(void) {
   vec4 diffuse, specular, ambient, totalColor;
   vec3 refl;
   float angleNL, angleVR;

   if (uMat.dColor.r <= -1.0 && uMat.dColor.g <= -1.0 && uMat.dColor.b <= -1.0)
      diffuse = texture2D(uTexUnit, vTexCoord);
   else
      diffuse = vec4(uMat.dColor, 1);
   if (uMat.dColor.r >= -1.0 && uMat.dColor.g >= -1.0 && uMat.dColor.b >= -1.0) {
      angleNL = clamp(dot(normalize(vNormal), normalize(uLightPos)), 0.0, 1.0);
      diffuse = diffuse * vec4(uLightColor, 1) * angleNL;
   }
   
   refl = reflect(-normalize(uLightPos), normalize(vNormal));
   angleVR = clamp(dot(normalize(refl), normalize(uViewPos-vPos)), 0.0, 1.0);
   specular = vec4(uMat.sColor, 1) * vec4(uLightColor, 1) * pow(angleVR, uMat.shine);
   
   ambient = vec4(uMat.aColor, 1) * vec4(uLightColor, 1);
   
   totalColor = diffuse + specular + ambient;
   gl_FragColor = vec4(totalColor.r, totalColor.g, totalColor.b, uMat.alpha);
}

