struct Material {
  vec3 aColor;
  vec3 dColor;
  vec3 sColor;
  float shine;
  float alpha;
};

attribute vec3 aPosition;
attribute vec2 aTexCoord;
attribute vec3 aNormal;

uniform mat4 uProjMatrix;
uniform mat4 uViewMatrix;
uniform mat4 uModelMatrix;
uniform vec3 uLightPos;
uniform vec3 uLightColor;
uniform vec3 uViewPos;
uniform Material uMat;

varying vec2 vTexCoord;
varying vec3 vNormal;
varying vec3 vPos;

void main() {
   vec4 vPosition, transPos, transNorm;

   // First model transforms
   vPosition = uModelMatrix * vec4(aPosition.x, aPosition.y, aPosition.z, 1);
   vPosition = uViewMatrix * vPosition;
   gl_Position = uProjMatrix * vPosition;

   // Apply transforms to position and normals
   transPos = uModelMatrix * vec4(aPosition, 1);
   transNorm = uModelMatrix * vec4(aNormal, 0);

   // Interpolate and send to fragment shader
   vTexCoord = aTexCoord;
   vPos = vec3(transPos);
   vNormal = vec3(transNorm);
}
